/*

  Nullsoft WASABI Source File License

  Copyright 1999-2001 Nullsoft, Inc.

    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


  Brennan Underwood
  brennan@nullsoft.com

*/

#include "../libpcm/libpcm.h"
#include "../libpcm/fader.h"
#include "../pversion.h"

#include "studio/services/svc_mediaconverter.h"
#include "studio/services/servicei.h"
#include "studio/corecb.h"
#include "studio/wac.h"
#include "common/rootcomp.h"
#include "attribs/cfgitemi.h"
#include "attribs/attrint.h"

#include "reader_wac.h"

class Xpcm : public svc_mediaConverterI
{
public:
  Xpcm();
  virtual ~Xpcm();
  // service
  static const char *getServiceName();
  virtual int canConvertFrom(svc_fileReader *reader, const char *name, const char *chunktype);
  virtual const char *getConverterTo();
  virtual int getInfos(MediaInfo *infos);
  virtual int processData(MediaInfo *infos, ChunkList *chunk_list, bool *killswitch);
  virtual int getLatency(void) { return 0; }
  // callbacks
  virtual int corecb_onSeeked(int newpos);
private:
  int seekpos;
  LIBPCM_DECODER *decoder;
  LIBPCM_FADER *fader;
  char samplebuffer[4096];
};

class WACcnv_libpcm : public WAComponentClient
{
public:
  WACcnv_libpcm();
  virtual ~WACcnv_libpcm();
  virtual GUID getGUID();
  virtual void onRegisterServices();
};


_int loop_count("loop count", 2);

Xpcm::Xpcm()
{
  decoder = 0;
  fader = 0;
  seekpos = -1;
}

Xpcm::~Xpcm()
{
  if (fader) libpcm_fader_terminate(fader);
  if (decoder) libpcm_close(decoder);
}

const char *Xpcm::getServiceName()
{
  return WAC_SERVICE_NAME;
}

const char *Xpcm::getConverterTo()
{
  return "PCM";
}

int Xpcm::canConvertFrom(svc_fileReader *reader, const char *name, const char *chunktype)
{
  return WAC_CAN_CONVERT_FROM(reader, name, chunktype);
}

int Xpcm::getInfos(MediaInfo *infos)
{
  LIBPCM_DECODER *inf_decoder;
  LIBPCM_FADER *inf_fader;
  svc_fileReader *reader = infos->getReader();
  if (!reader) return 0;
  libpcm_IReader *plibpcm_reader = libpcm_CreateWacReader(reader, Std::filename(infos->getFilename()));
  if (!plibpcm_reader) return 0;
  inf_decoder = libpcm_open_from_stream(plibpcm_reader);
  plibpcm_reader->lpVtbl->Release(plibpcm_reader);
  if (inf_decoder)
  {
    inf_fader = libpcm_fader_initialize(inf_decoder);
    if (inf_fader)
    {
      libpcm_fader_configure(inf_fader, loop_count, 5000);
      infos->setLength(libpcm_fader_get_length_ms(inf_fader));
      infos->setTitle(libpcm_get_title(inf_decoder));
      infos->setInfo(StringPrintf("%ikbps %ibit %ihz %s", (libpcm_get_bitrate(inf_decoder) / 1000), libpcm_get_bitspersample(inf_decoder), libpcm_get_samplerate(inf_decoder), (libpcm_get_numberofchannels(inf_decoder) == 2) ? "stereo" : "mono"));
      libpcm_fader_terminate(inf_fader);
    }
    libpcm_close(inf_decoder);
  }
  return 0;
}

int Xpcm::processData(MediaInfo *infos, ChunkList *chunk_list, bool *killswitch)
{
  if (fader == 0)
  {
    svc_fileReader *reader = infos->getReader();
    if (!reader) return 0;
    libpcm_IReader *plibpcm_reader = libpcm_CreateWacReader(reader, Std::filename(infos->getFilename()));
    if (!plibpcm_reader) return 0;
    decoder = libpcm_open_from_stream(plibpcm_reader);
    plibpcm_reader->lpVtbl->Release(plibpcm_reader);
    if (!decoder) return 0;
    fader = libpcm_fader_initialize(decoder);
    if (!fader) return 0;
    libpcm_fader_configure(fader, loop_count, 5000);
  }

  if (seekpos != -1)
  {
    libpcm_seek_ms(decoder, seekpos);
    seekpos = -1;
  }

  int eof = 0;
  unsigned ba = libpcm_get_blockalign(decoder);
  unsigned lens = sizeof(samplebuffer) / ba;
  unsigned len = libpcm_fader_read(fader, samplebuffer, lens);
  if (len < lens) eof = 1;
  len *= ba;
  if (len == 0) return 0;

  ChunkInfosI *ci = new ChunkInfosI();
  ci->addInfo("srate", libpcm_get_samplerate(decoder));
  ci->addInfo("bps", libpcm_get_bitspersample(decoder));
  ci->addInfo("nch", libpcm_get_numberofchannels(decoder));

  chunk_list->setChunk("PCM", samplebuffer, len, ci);

  return !eof;
}

int Xpcm::corecb_onSeeked(int newpos)
{
  seekpos = newpos;
  return 0;
}

static WACcnv_libpcm wac;
WAComponentClient *the = &wac;

// {83DBF600-DBE6-11d6-B153-0090CC1DDA59}
static const GUID guid = 
{ 0x83dbf600, 0xdbe6, 0x11d6, { 0xb1, 0x53, 0x0, 0x90, 0xcc, 0x1d, 0xda, 0x59 } };

WACcnv_libpcm::WACcnv_libpcm() : WAComponentClient(WAC_CLIENT_NAME) {
  registerService(new waServiceT<svc_mediaConverter, Xpcm>);
}

WACcnv_libpcm::~WACcnv_libpcm() {
}

GUID WACcnv_libpcm::getGUID() {
  return guid;
}

void WACcnv_libpcm::onRegisterServices() {
  WAC_REGISTER_EXTS
  registerAttribute(&loop_count);
}

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker,"/OPT:NOWIN98")
//#pragma comment(linker,"/ENTRY:DllMain")
#pragma comment(linker,"/MERGE:.rdata=.text")
#pragma comment(linker,"/STUB:../stub.exe")
#endif

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		//DisableThreadLibraryCalls(hinstDLL);
	}
	return TRUE;
}
